package usherbrooke.ift630;

import java.util.ArrayList;
import java.util.concurrent.BlockingQueue;

public class Bus extends Thread {
	private ArrayList<Passenger>					passengers;
	private ArrayList<Stop>							stops;
	private String									name;
	private int										id;
	private int										maxCapacity;
	private int										currentCapacity;
	private int										timeBetweenStops;
	private int										timeToNextStop;
	private int										timeEmbark;
	private Stop									nextStop;
	private BlockingQueue<BusInformationMessage>	blockingQueue;

	// returns time for a single passenger to embark.
	private int timeEmbark() {
		return (int) (Math.random() * timeEmbark);
	}

	// set time between stop.
	private int calculateTimeBetweenStops() {
		timeToNextStop = (int) ((timeBetweenStops / 3) +  (Math.random() * 2 * timeBetweenStops / 3));
		return timeToNextStop;
	}

	// check if a passenger must leave at next stop.
	private boolean stopAsked() {
		boolean res = false;
		for (Passenger p : passengers){
			if (p.getDest() == nextStop) {
				res = true;
				Logger.getInstance().print(id, "[BUS] " + name + " stop asked!");
				break;
			}
		}

		return res;
	}

	// travel to next stop. Waits the travel time.
	private Stop goToNextStop() throws InterruptedException, IndexOutOfBoundsException {
		// if no reason to stop, skip current stop
		boolean nextStopEmpty, overMaxCapacity;
		do {
			nextStop = stops.remove(0);

			sendNextStopInfo();

			waitStop();

			overMaxCapacity = currentCapacity >= maxCapacity;
			nextStopEmpty = getNextStopPassengers().isEmpty();

			if (overMaxCapacity)
				Logger.getInstance().print(id, "[BUS] " + name + " im full, sorry!");
			if (nextStopEmpty && !stopAsked())
				Logger.getInstance().print(id, "[BUS] " + name + " next " + nextStop.getStopName() + " is empty!");
		} while (!stopAsked() && (nextStopEmpty || overMaxCapacity));

		return nextStop;
	}

	// send Bus info msg to next stop
	private void sendNextStopInfo() throws InterruptedException{
		BusInformationMessage info = new BusInformationMessage(this, nextStop, timeToNextStop);
		synchronized(blockingQueue) {
			blockingQueue.put(info);
		}
	}

	// get all passenger waiting at next stop
	// useful for knowing if we should stop
	private ArrayList<Passenger> getNextStopPassengers() {
		ArrayList<Passenger> res = new ArrayList<Passenger>();

		if (nextStop != null)
			for (Passenger p : passengers) {
				if (p.getDest() == nextStop) {
					res.add(p);
				}

			}

		return res;
	}

	// lock until passenger got in
	private void waitEmbark() throws InterruptedException {
		synchronized (this) {
			sleep(timeEmbark());
		}
	}

	// lock until passenger got out
	private void waitStop() throws InterruptedException {
		synchronized (this) {
			sleep(timeToNextStop);
		}
	}

	@Override
	public void run() {
		try {
			// run until terminus
			while (nextStop != null) {
				// new stop !
				Logger.getInstance().print(id, "[BUS] " + name + " reached stop " + nextStop.getStopName() + "!");

				disembarkPassengers();
				embarkPassengers();
				calculateTimeBetweenStops();

				try {
					nextStop = goToNextStop();
				} catch (IndexOutOfBoundsException e) {
					nextStop = null;
				}
			}
		} catch (InterruptedException e) {
			Logger.getInstance().print(id, "[BUS] exception: " + e.toString());
		}
		Logger.getInstance().print(id, "[BUS] " + name + " exiting!");
	}

	Bus(ArrayList<Stop> s, int id, int timeStop, int timeEmbark, int maxCapacity) {
		this.timeBetweenStops = timeStop * 1000 + 1;
		this.timeEmbark =		timeEmbark * 1000 + 1;
		this.name =				"Bus n°" + id;
		this.id =				id;
		this.passengers =		new ArrayList<Passenger>();
		this.stops =			s;
		this.nextStop =			stops.get(0);
		this.currentCapacity =	0;
		this.maxCapacity =		maxCapacity;
	}

	// disembark passenger at a stop. Synchronizes, locks on that stop, and waits.
	public synchronized void disembarkPassengers() throws InterruptedException {
		// lock on next stop
		synchronized (nextStop) {
			for (Passenger p : passengers) {
				waitEmbark();

				Logger.getInstance().print(id,
						"\t[DISEMBARK] " + p.getName() + " at " + nextStop.getStopName());

				if (p.getDest() == nextStop) {
					p.disembark();
					currentCapacity--;
				}
			}
		}
	}

	// embark passenger at a stop. Synchronizes, locks on that stop, and waits.
	// check for overflow, and if passenger should embark in the first place
	public synchronized void embarkPassengers() throws InterruptedException {
		synchronized (nextStop) {
			try {
				ArrayList<Passenger> list = nextStop.getPassengerByDest(stops);
				for (Passenger p : list) {
					waitEmbark();
					if (currentCapacity >= maxCapacity)
						continue;
					passengers.add(p);
					nextStop.removePassenger(p);
					currentCapacity++;

					Logger.getInstance().print(id,
							"\t[EMBARK] " + p.getName() + " at " + nextStop.getStopName());
				}
			} catch (IndexOutOfBoundsException e) {
				Logger.getInstance().print(id, "Exception: " + e.getMessage());
			}
		}
	}

	// pretty print!
	public void printDetails(int indent) {
		Logger.getInstance().print(id, "\t".repeat(indent) + "---".repeat(3) + " Bus details " + "---".repeat(3));
		Logger.getInstance().print(id, "\t".repeat(indent) + name);
		Logger.getInstance().print(id, "\t".repeat(indent) + "time between stops: " + timeBetweenStops);
		Logger.getInstance().print(id, "\t".repeat(indent) + "max capacity: " + maxCapacity);
		Logger.getInstance().print(id, "\t".repeat(indent) + "current stop: " + nextStop.getStopName());
		Logger.getInstance().print(id, "\t".repeat(indent) + "stops: ");


		if (nextStop != stops.get(0))
			nextStop.printDetails(id, indent + 1);
		for (Stop s : stops) {
			s.printDetails(id, indent + 1);
		}
	}

	// pretty print w/ 0 indent
	public void printDetails() {
		printDetails(0);
	}

	// return stop list
	public ArrayList<Stop> getStops() {
		ArrayList<Stop> res = new ArrayList<Stop>(stops);
		if (nextStop != stops.get(0))
			res.add(0, nextStop);

		return res;
	}

	// return last stop
	public Stop getTerminus() {
		return stops.get(stops.size() - 1);
	}

	public String getNameBus() {
		return name;
	}

	public void setBlockingQueue(BlockingQueue<BusInformationMessage> q) {
		blockingQueue = q;
	}
}
